# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.API.Game do
  @moduledoc """
  This module includes the interface for defining and registering
  a game with Amethyst.
  """
  @callback instantiate() :: {:ok, new_state :: term} | {:error, reason :: term}
  @callback login(connection_process :: pid(), player_cfg :: keyword(), state :: term) :: {:ok, new_state :: term} | {:refuse, new_state :: term}

  defmacro __using__(opts) do
    meta = Keyword.get(opts, :meta, [])
    quote do
      @behaviour Amethyst.API.Game
        def child_spec(state) do
          %{
            id: __MODULE__,
            start: {__MODULE__, :register_self, []}
          }
        end
        def register_self() do
          Amethyst.GameRegistry.register(__MODULE__, unquote(meta))
          Process.sleep(:infinity)
        end
        def start(state) do
          loop(state)
        end
        defp loop(state) do
          receive do
            {:login, caller, cfg} ->
              case login(caller, cfg, state) do
                {:ok, state} ->
                  send(caller, :ok)
                  loop(state)
                {:refuse, state} ->
                  send(caller, :refuse)
                  loop(state)
              end
          end
        end
    end
  end

  def login(pid, cfg) do
    send(pid, {:login, self(), cfg})
    receive do
      :ok -> :ok
      :refuse -> :refuse
    end
  end
end
